#ifndef NEWTON_HH
#define NEWTON_HH

#include "funclass.hh"
#include <cmath>
#include <stdexcept>      // std::exception

template <typename FUNC, typename T, typename INT >
T // return the value of the approximated root
newton( FUNC const & f,        // instance of object function class
        T    const & x0,       // initial guess
        INT          maxIter,  // maximum number of iteration
        T    const & tolerance // tolereance used to stop iterations
       ) {
  T x = x0 ;
  for ( INT i = 0 ; i < maxIter ; ++i ) {
    T fun  = f(x) ;
    T Dfun = f.D(x) ;
    if ( std::abs(Dfun) < 1E-10*std::abs(fun) )
      throw std::runtime_error("newton, divide by (near) 0") ;
    T h = fun/Dfun ;
    x -= h ;
    if ( std::abs(h) < tolerance ) return x ; // all ok!

    if ( std::abs(x-x0) > 1E10 )
      throw std::runtime_error("newton is diverging") ;
  }
  throw std::runtime_error("newton not converged within iteration limits") ;
} ;

#endif
